#include "std.h"
#include "malical.h"
#include "mclstdlib.h"

/////////////////////////////////
//     Operator of Object
///////////////////////////////// 

void obj_ini (Object *d)
{
	d->value.str = 0;
	d->tag = VT_NIL;
}
void obj_clear (Object *d)
{
	switch (d->tag)
	{
	
	
		case VT_ARR:
			d->value.ref = 0;
			break;
		case VT_REF:
			d->value.ref = 0;
			break;
		case VT_NUM:
			d->value.d = 0.0;
			break;
		case VT_STR:
			if (d->value.str!=NULL) free(d->value.str);
			d->value.str = NULL;
			break;
		default:break;
	}
}

Object* obj_copy (Object *d,const Object *s)
{
	obj_clear(d);
	d->tag = s->tag ;
	switch (s->tag)
	{
		case VT_ARR:
			d->value.ref = s->value.ref;
			break;
		case VT_REF:
			d->value.ref = s->value.ref;
			break;
		case VT_NUM:
			d->value.d = s->value.d;
			break;
		case VT_NIL:
			d->value.d = s->value.d;
			break;
		case VT_STR:
		{
			int len = strlen(s->value.str)+1;
			d->value.str = (char*)malloc(len);
			strcpy(d->value.str,s->value.str);
		}
		default:break;
	}
	return d;
}
Object * obj_create (Object *d,Type tag,void *value)
{
	obj_clear(d);
	d->tag = tag;
	switch (tag)
	{
		case VT_ARR:
			d->value.ref = value;
			break;
		case VT_REF:
			d->value.ref = value;
			break;
		case VT_NIL:
			d->value.d = 0;
			break;
		case VT_NUM:
			d->value.d = *(double*)value;
			break;
		case VT_STR:
		{
			char *str = (char*)value;
			int len = strlen(str)+1;
			d->value.str = (char*)malloc(len);
			strcpy(d->value.str,str);
			break;
		}
		default:break;
	}
	return d;
}
Object * obj_ref_to_obj (Object *o)
{
	if (o->tag == VT_REF)
	{
		Object *v_o;
		v_o = (Object*)(o->value.ref);
		return v_o;
	}
	else
		return 0;
}

char * obj_to_str(Object *s, char *str)
{
	switch (s->tag)
	{
		case VT_ARR:
			strcpy(str,"Array");
			break;			
		case VT_REF:
			strcpy(str,"$");
			break;
		case VT_NIL:
			strcpy(str,"Nil");
			break;
		case VT_NUM:
			sprintf(str,"%lf",s->value.d);
			break;
		case VT_STR:
			strcpy(str,s->value.str);
			break;
		default:break;
	}
	return str;
}
void objs_push (ObjectStack *s,Object *n_o)
{
	s->stk[s->size] = (Object*)malloc(sizeof(Object));
	obj_ini(s->stk[s->size]);
	if (n_o != 0) obj_copy(s->stk[s->size],n_o);
	s->size++;
}
void objs_pop (ObjectStack *s,Object *n_o)
{
	s->size--;
	if (n_o != 0) obj_copy(n_o,s->stk[s->size]);
	obj_clear(s->stk[s->size]);
	free(s->stk[s->size]);
}
void objs_clear (ObjectStack *s)
{
	while(s->size>0)
	{
		s->size--;
		obj_clear(s->stk[s->size]);
		free(s->stk[s->size]);
	}
}
/////////////////////////////////
//      List of Operator
///////////////////////////////// 

SIGNVALUE SignValue[]=
{
	{"+",	SIGN_ADD},
	{"-",	SIGN_SUB},
	{"*",	SIGN_MUL},
	{"/",	SIGN_DIV},
	{"%",	SIGN_MOD},
	{"^",	SIGN_POW},
	{"(",	SIGN_BLK},
	{")",	SIGN_BRK},
	{"and",	SIGN_AND},
	{"or",	SIGN_OR	},
	{"not",	SIGN_NOT},
	{"==",	SIGN_EQ },
	{"!=",	SIGN_NEQ},
	{">",	SIGN_BT },
	{">=",	SIGN_BE },
	{"<",	SIGN_LT },
	{"<=",	SIGN_LE },
	{"=",	SIGN_STO}
};
const int SignSize = sizeof(SignValue)/sizeof(SIGNVALUE);

int GetPerendence (SIGN s)
{
	switch (s)
	{
		case SIGN_STO:return -99;// =
		case SIGN_OR :return -4;// or
		case SIGN_AND:return -3;// and
		case SIGN_EQ :			// ==
		case SIGN_NEQ:return -2;// !=
		case SIGN_BT :			// >
		case SIGN_BE :			// >=
		case SIGN_LT :			// <
		case SIGN_LE :return -1;// <=
		case SIGN_ADD:
		case SIGN_SUB:return 0;
		case SIGN_MUL:
		case SIGN_DIV:
		case SIGN_MOD:return 1;
		case SIGN_POW:return 2;
		case SIGN_BLK:
		case SIGN_BRK:return 99;
	}
	return -999;
}
////  string is operator?  
int SignToValue (char *sign)
{
	int i;
	for (i=0;i<SignSize;++i)
		if (str_eq(SignValue[i].name,sign)) return SignValue[i].value;
	return -1;
}

////  string is cfunction?  
PCFUNC FuncToPointer (char *fun)
{
	int left, right, mid, cmp;
	left = 0;
	right = CfuncSize - 1;

	while( left <= right ){
		mid = (left + right) / 2;
		cmp = strcmp( fun, CFunction[mid].name );
		if( cmp < 0) 
			right = mid - 1;
		else if( cmp > 0 )
			left = mid + 1;
		else
			return CFunction[mid].f;
	}
	return 0;
}

///// Eval RL
int EvalLR (Object *l,Object *r,SIGN sign,Object *result)
{
	double d;

	if (o_is_ref(r)) r = obj_ref_to_obj(r);
	if (l!=0) if (o_is_ref(l) && sign != SIGN_STO) l = obj_ref_to_obj(l);

	if (o_is_num(l) && o_is_num(r))
	{
		switch (sign)
		{
			case SIGN_ADD:
				d = l->value.d + r->value.d;goto make_results;
			case SIGN_SUB:
				d = l->value.d - r->value.d;goto make_results;
			case SIGN_MUL:
				d = l->value.d * r->value.d;goto make_results;
			case SIGN_DIV:
				d = l->value.d / r->value.d;goto make_results;

			case SIGN_EQ :
				d = l->value.d == r->value.d;goto make_results;
			case SIGN_NEQ:
				d = l->value.d != r->value.d;goto make_results;
			case SIGN_BT :
				d = l->value.d >  r->value.d;goto make_results;
			case SIGN_BE :
				d = l->value.d >= r->value.d;goto make_results;
			case SIGN_LT :
				d = l->value.d <  r->value.d;goto make_results;
			case SIGN_LE :
				d = l->value.d <= r->value.d;goto make_results;
			case SIGN_AND:
				d = l->value.d && r->value.d;goto make_results;
			case SIGN_OR :
				d = l->value.d || r->value.d;goto make_results;

			case SIGN_MOD:
				d = (int)l->value.d % (int)r->value.d;goto make_results;
			case SIGN_POW:
				d = pow(l->value.d,r->value.d);goto make_results;
			default:
				MalicalTypeError(l->tag,r->tag,sign);
				break;
		}
	}
	else if (o_is_str(l) && o_is_str(r))
	{
		switch(sign)
		{
			case SIGN_EQ :
				d = str_eq(l->value.str,r->value.str);goto make_results;
			case SIGN_NEQ:
				d =!str_eq(l->value.str,r->value.str);goto make_results;
			case SIGN_ADD:
				{
					char buf[TOKEN_BUFFER_SIZE];
					strcpy(buf,l->value.str);
					strcat(buf,r->value.str);
					obj_create (result,VT_STR,buf);
				}
				break;
			default:
				MalicalTypeError(l->tag,r->tag,sign);
				break;
		}
	}
	else if (l->tag==VT_NIL && o_is_num(r) && sign==SIGN_NOT)
	{	
		d = !r->value.d;
		goto make_results;
	}
	else if (l->tag == VT_REF && sign==SIGN_STO)
	{
		Object * v_o = (Object*)l->value.ref;
		
		obj_copy(v_o,r);
		obj_copy(result,r);
	}
	else if (l->tag != r->tag)
	{
		switch (sign)
		{
			case SIGN_EQ :
				d = 0;goto make_results;
			case SIGN_NEQ:
				d = 1;goto make_results;
			default:
				MalicalTypeError(l->tag,r->tag,sign);
				break;
		}
make_results:
			obj_create (result,VT_NUM,&d);
	}
	return 0;
}
///// Eval Main
int EvalExpr (char *expr,Object *objr)
{
	int pl = 0,type;
	ObjectStack obj_stk={0,0};
	SIGN sstk[STACK_SIZE];int stop=0;
	while (expr[pl])
	{
		pl = GetToken(expr,pl,&type);
		if (pl==-1)break;
		if (type==TT_NUM || type==TT_STR || type==TT_NIL)
		{
			Object temp_obj;
			double d;
			obj_ini(&temp_obj);
			switch(type)
			{
				case TT_NIL:
					d = 0;
					obj_create(&temp_obj,VT_NIL,&d);
					break;
				case TT_NUM:
					d = atof(token);
					obj_create(&temp_obj,VT_NUM,&d);
					break;
				case TT_STR:
					obj_create(&temp_obj,VT_STR,token);
					break;
				default:break;
			}
			objs_push(&obj_stk,&temp_obj);
		}
		else if (type==TT_BRACE_L)
		{
			int r;
			Object o;
			char expr_buf[TOKEN_BUFFER_SIZE];
			Array *a;
			obj_ini(&o);
			a = array_new();
			while ((r=GetTokenE(expr,'[',']',pl))!=-1)
			{
				strcpy(expr_buf,token);
				EvalExpr(expr_buf,&o);
				array_push_back (a,&o);
				pl = r;
			}
			pl = Match(expr,"}",0,pl);
			objs_push(&obj_stk,0);
			obj_create(obj_stk.stk[obj_stk.size-1],VT_ARR,a);
		}
		else if (type==TT_BK_L)
		{
			char expr_buf[TOKEN_BUFFER_SIZE];
			int r;
			Object o;
			Object oa;
			Element *e;
			
			obj_ini(&o);
			obj_ini(&oa);
			pl = GetTokenE(expr,'[',']',pl-1);
			strcpy(expr_buf,token);
			EvalExpr(expr_buf,&o);
			if (!o_is_num(&o)) MalicalError("Array index must be num");
			r = o.value.d;
			obj_clear (&o);
			objs_pop(&obj_stk,&oa);
			if (o_is_arr(&oa))
			{
				e = array_element((Array*)(oa.value.ref),r);
			}
			else if (o_is_ref(&oa))
			{
				Object *o_ref_arr=obj_ref_to_obj(&oa); 
				if (o_is_arr(o_ref_arr))
				{
					e = array_element((Array*)(o_ref_arr->value.ref),r);
				}else MalicalError("No array!");
			}
			else MalicalError("No array!");
			if (e)
				obj_create (&o,VT_REF,&e->obj);
			else
				obj_create (&o,VT_NIL,0);
			obj_clear(&oa);
			objs_push  (&obj_stk, &o);
		}
		else if (type==TT_SIGN)
		{
			int sign;
			sign = SignToValue(token);
			if (sign==-1)
			{
				MalicalTokenError();
			}
			if (sign==SIGN_BRK)
			{
				while (stop>0)
				{
					Object l,r;
					SIGN p_s;
					obj_ini(&l);
					obj_ini(&r);
					p_s = sstk[--stop];
					if (p_s==SIGN_BLK) break;
					objs_pop(&obj_stk,&r);
					if (p_s!=SIGN_NOT)objs_pop(&obj_stk,&l);	
					objs_push(&obj_stk,0);
					EvalLR(&l,&r,p_s,obj_stk.stk[obj_stk.size-1]);
					obj_clear(&l);
					obj_clear(&r);
				}
			}
			else if (GetPerendence(sign) <= GetPerendence(sstk[stop-1]))
			{
				if (sstk[stop-1]!=SIGN_BLK && sstk[stop-1]!=SIGN_STO)
				{
					Object l,r;
					SIGN p_s;
					obj_ini(&l);
					obj_ini(&r);
					p_s = sstk[--stop];
					if (p_s==SIGN_BLK) break;
					objs_pop(&obj_stk,&r);
					if (p_s!=SIGN_NOT)objs_pop(&obj_stk,&l);	
					objs_push(&obj_stk,0);
					EvalLR(&l,&r,p_s,obj_stk.stk[obj_stk.size-1]);
					obj_clear(&l);
					obj_clear(&r);
				}
				sstk[stop++] = sign;
			}
			else
			{
				sstk[stop++] = sign;
			}
		}
		else if (type==TT_SYM)
		{
			int r;
			PCFUNC pcfunc;
			Variable *pv;
			int posf;
			//// funcion call
			pcfunc = FuncToPointer(token);
			if (pcfunc!=0)
			{
				ObjectStack arg_stk={0,0};
				char expr_buf[TOKEN_BUFFER_SIZE];
				while ((r=GetTokenE(expr,'[',']',pl))!=-1)
				{
					strcpy(expr_buf,token);
					objs_push(&arg_stk,0);
					EvalExpr(expr_buf,arg_stk.stk[arg_stk.size-1]);
					pl = r;
				}
				objs_push(&obj_stk,0);
				pcfunc(&arg_stk,obj_stk.stk[obj_stk.size-1]);
				objs_clear(&arg_stk);
				continue;
			}
			///// variabel value
			pv = find_var(token);
			if (pv != 0)
			{
				objs_push(&obj_stk,0);
				obj_create(obj_stk.stk[obj_stk.size-1],VT_REF,&pv->obj);
				continue;
			}
			///// user func
			posf = find_func(token);
			if (posf>0)
			{
				ObjectStack arg_stk={0,0};
				char expr_buf[TOKEN_BUFFER_SIZE],fun_name[NAME_LENGTH];
				strcpy(fun_name,token);
				while ((r=GetTokenE(expr,'[',']',pl))!=-1)
				{
					strcpy(expr_buf,token);
					objs_push(&arg_stk,0);
					EvalExpr(expr_buf,arg_stk.stk[arg_stk.size-1]);
					pl = r;
				}
				CallFun(fun_name,&arg_stk);
				objs_push(&obj_stk,&ReturnObject);
				objs_clear(&arg_stk);
				continue;
			}
			MalicalTokenError();
		}
		else
			MalicalTokenError();
	}
	while (stop>0)
	{
		Object l,r;
		SIGN p_s;
		obj_ini(&l);
		obj_ini(&r);
		p_s = sstk[--stop];
		objs_pop(&obj_stk,&r);
		if (p_s!=SIGN_NOT)objs_pop(&obj_stk,&l);	
		objs_push(&obj_stk,0);
		EvalLR(&l,&r,p_s,obj_stk.stk[obj_stk.size-1]);
		obj_clear(&l);
		obj_clear(&r);
	}
	if (obj_stk.size!=1)
	{
		MalicalError("Eval error");
		objs_clear(&obj_stk);
		return -1;
	}
	if(objr != 0)
	{
		objs_pop(&obj_stk,objr);		
		if (o_is_ref(objr))
		{
			obj_copy(objr,obj_ref_to_obj(objr));
		}
	}
	objs_clear(&obj_stk);
	return 1;
}